/***************************************************************************
 *   Copyright (C) 2007 by Martin Apel,,,                                  *
 *   kermit@frog                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "kausn.h"

#include <iostream>


using namespace std;

void KausN::Print(SubSet& sSet)
{
  SubSet::iterator i;

  for (i=sSet.begin(); i != sSet.end(); ++i)
  {
    cerr << *i << ",";
  }
  cerr << endl;
}

KausN::KausN(unsigned k, unsigned n) : VCryptScheme(n),
                                       m_uiNeeded(k)
{
  CreateBaseMatrices();
}


KausN::~KausN()
{
}

//this function gets the next configuration of the column
//it guarantees, that the number of ones in the column stay
//the same
bool KausN::NextCol(Column& S)
{
  int pos = S.size() - 1;
  unsigned uiToSet=0;
  while(1)
  {
    //find last 1
    while(S[pos]!=1)
    {
      pos--;
      if (pos < 0)
        return false; //no 1 left to remove
    }
    //remove found 1
    S[pos]=0;
    uiToSet++;

    //try to set the ones from here on
    if (pos < static_cast<int>((S.size() - uiToSet)))
    {
      //This will work - set the 1s
      for (pos++; uiToSet>0; uiToSet--, pos++)
        S[pos]=1;
      return true;
    }
    //if setting the 1s would not be succesful, remove next 1 - that means start at the top
  }
  //cannot be reached
  return false;
}

void KausN::Add(unsigned p, DMatrix** S)
{
  Column cCol(m_uiUser);
  unsigned q;

  //compute minimal number of ones that are needed
  //(look at Drostes Paper for an explanation)
  if (p > (m_uiNeeded-p))
    q = p + (m_uiUser-m_uiNeeded);
  else
    q = p;

  /*********** create new columns  START *******************/

  //Startkonfiguration setzen
  for (unsigned i=0;i < q;i++)
    cCol[i]=1;

  do
  {
    //if matrix is empty - create matrix with one column
    if (*S==NULL)
    {
      (*S) = new DMatrix(1, m_uiUser);
      (*S)->SetCol(0, cCol);
    }
    else
      (*S)->InsertCol((*S)->GetXDim(), cCol); //Einfuegen immer am Ende
  } while(NextCol(cCol));
  /*********** create new columns END **********************/
}

//this function finds an unbalanced column and returns the number of 1s in that
//column
//positive value: number of 1s in found unbalanced column of S1 (add to S0)
//negative value: negative number of 1s in found unbalanced column of S0 (add to S1)
bool KausN::FindUnbalanced(DMatrix* S0, DMatrix* S1, set<unsigned>& sS0Matched,
                           set<unsigned>& sS1Matched, bool& bS0, unsigned& uiOnesRet)
{
  int iEl1;
  unsigned m0, m1;  //Sizes of S0 and S1

  //determine sizes
  m0 = S0->GetXDim();
  m1 = S1->GetXDim();

  //all columns not marked as balanced have to be the same in S0 and S1
  //Try to match columns
  for (unsigned i = 0; i < m0; i++)
  {
    //if this column has already been balanced - then continue with next column
    if (sS0Matched.count(i) > 0)
      continue;

    //find unbalanced column of s1
    for (unsigned c = 0; c < m1; c++)
    {
      if (sS1Matched.count(c) > 0)
        continue;
      //if the columns match
      if (MatchCols(S0, i, S1, c, m_uiNeeded))
      { //then mark them as matched
        sS0Matched.insert(i);
        sS1Matched.insert(c);
        break;  //return to outer loop
      }
    }

    //Have our column of s0 been matched?
    //if not - we found an unbalanced column
    if (sS0Matched.count(i) == 0)
    {
      uiOnesRet = 0;
      bS0 = true; //unbalanced column found in S0
      //Count ones
      for (unsigned c = 0; c < m_uiNeeded; c++)
      {
        S0->GetElem(i, c, iEl1);

        if (iEl1 == 1)
          uiOnesRet++;
      }

      //return
      return true; //unbalanced column found
    }
  }

  //at this point - all columns of S0 are either balanced or matched
  //if there is at least one unbalanced/unmatched column in S1 left
  //this is the one to return
  for (unsigned i = 0; i < m1; i++)
  {
    if (sS1Matched.count(i)==0)
    {
      bS0 = false; //unbalanced column was found in S1
      //Count ones
      uiOnesRet = 0;
      for (unsigned c = 0; c < m_uiNeeded; c++)
      {
        S1->GetElem(i, c, iEl1);
        if (iEl1 == 1)
          uiOnesRet++;
      }

      return true; //unbalanced column found
    }
  }
  //everything in s1 was matched too!!
  //thats great - we finished it!!

  return false; //no unbalanced column found
}

void KausN::CreateBaseMatrices()
{
  //The Base Matrices
  DMatrix* S0 = NULL;
  DMatrix* S1 = NULL;
  //These sets hold the columns that are balanced or have been matched
  set<unsigned> sS0Matched;
  set<unsigned> sS1Matched;

  bool bS0;
  unsigned uiOnes;
  int iRet;

  //Initialize base matrices
  for (unsigned p = 0; p <= m_uiNeeded; p++)
  {
    if (p%2) //odd
      Add(p, &S1);
    else
      Add(p, &S0);
  }

  //find balanced columns
  //every column (of size k resp. m_uiNeeded) with an even number of ones has to be
  //present in S0 - and is marked as balanced (it is important, that every column is
  //only matched once!!)

  //build column
  Column cTestCol(m_uiNeeded, 0);
  Column cTestCol2;
  uiOnes = 0;
  do
  {
    if (uiOnes % 2) //odd number of ones
    { //Search in S1
      for (unsigned c = 0; c < S1->GetXDim(); c++)
      {
        cTestCol2 = S1->GetCol(c, iRet);
        if (Matched(cTestCol, cTestCol2))
        {
          sS1Matched.insert(c);
          break; //Back to do-loop
        }
      }
      //In the loop above it is not necessary to check if a column has been
      //matched at all, because every column has to be matched, by design of
      //the algorithm
    }
    else //Even number of ones
    { //Search in S0
      for (unsigned c = 0; c < S0->GetXDim(); c++)
      {
        cTestCol2 = S0->GetCol(c, iRet);
        if (Matched(cTestCol, cTestCol2))
        {
          sS0Matched.insert(c);
          break; //Back to do-loop
        }
      }
    }
  } while (NextBitString(cTestCol, uiOnes));

  //Match remaining columns
  while (FindUnbalanced(S0, S1, sS0Matched, sS1Matched, bS0, uiOnes))
  {
    //if S0 had unmatched columns - add to S1
    if (bS0)
    {
      Add(uiOnes, &S1);
    }
    //if S1 had unmatched columns - add to S0
    else
    {
      Add(uiOnes, &S0);
    }
  }

  m_pcWhiteBase = S0;
  m_pcBlackBase = S1;
}

bool KausN::Matched(const Column& c1, const Column& c2)
{
  Column::const_iterator i1, i2;
  i1 = c1.begin();
  i2 = c2.begin();

  while ((i1 != c1.end()) && (i2 != c2.end()))
  {
    if (*i1 != *i2)
      return false;
    ++i1;
    ++i2;
  }

  return true;
}

bool KausN::NextBitString(Column& cCol, unsigned& uiOnes)
{
  unsigned toAdd = 1;

  uiOnes = 0;
  for (unsigned i = 0; i < cCol.size(); i++)
  {
    //add
    if (toAdd == 1)
    {
      if (cCol[i] == 1)
        cCol[i] = 0; //toAdd bleibt 1 - da ein uebertrag entsteht
      else
      {
        cCol[i] = 1;
        toAdd   = 0;
      }
    }
    //count 1s
    if (cCol[i] == 1)
      uiOnes++;
  }

  //if overflowed - output false
  if (toAdd > 0)
    return false;
  else
    return true;
}
