/***************************************************************************
 *   Copyright (C) 2007 by Martin Apel,,,                                  *
 *   kermit@frog                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "vcryptscheme.h"

#include <cmath>
#include <QtGui>

#include <iostream>

using namespace std;

VCryptScheme::VCryptScheme(unsigned uiUser) : m_uiUser(uiUser)
{
  m_pcBlackBase = m_pcWhiteBase = NULL;
  m_sBaseName   = "";
  m_sSourceFile = "";
  m_uiSubPixelX = m_uiSubPixelY = 0;
}


VCryptScheme::~VCryptScheme()
{
  delete m_pcBlackBase;
  delete m_pcWhiteBase;
}

DMatrix* VCryptScheme::TranslatePixel(QRgb cColor)
{
  bool bBlack;

  //BaseMatrices valid?
  if (m_pcBlackBase == NULL || m_pcWhiteBase == NULL)
    return NULL;

  //Examine Color
  if (qGreen(cColor) == 0 && qBlue(cColor) == 0 && qRed(cColor) == 0)
    bBlack = true;
  else
    bBlack = false;

  //get random permutation of base matrix
  if (bBlack)
    return m_pcBlackBase->Permute();
  else
    return m_pcWhiteBase->Permute();
}

int VCryptScheme::GetBaseMatrices(DMatrix** pcWhiteBase, DMatrix** pcBlackBase)
{
  //if the base matrices aren't valid - return error
  if (m_pcWhiteBase == NULL || m_pcBlackBase == NULL)
    return 0;

  //Save pointers of base matrices to given locations
  *pcWhiteBase = m_pcWhiteBase;
  *pcBlackBase = m_pcBlackBase;

  //everything is ok
  return 1;
}


QStringList* VCryptScheme::Share(bool bForceSquare)
{
  QStringList* slRet = new QStringList;
  QImage** pcShares;
  QImage*  pcSecret;
  QSize    cSecretSize;
  QSize    cSubPixel;
  QSize    cShareSize;
  DMatrix* pcPixel;
  unsigned x,y;
  int iRet;

  //open source file
  pcSecret = new QImage(m_sSourceFile.c_str());
  //todo: open succesful?

  cSecretSize = pcSecret->size();

  //Create the array for the shares
  pcShares = new QImage*[m_uiUser];

  //delete old base matrices
  delete m_pcBlackBase;
  delete m_pcWhiteBase;
  m_pcBlackBase = m_pcWhiteBase = NULL;

  //Get Base matrices
  CreateBaseMatrices();
  Optimize();

  //Extend Base matrices (to achieve m=x*y)
  FindOptimalExpansion(m_pcWhiteBase->GetXDim(), x, y, bForceSquare);
  ModifyBaseMatrices(x, y, m_pcWhiteBase, m_pcBlackBase);
  cSubPixel.setWidth(x);
  cSubPixel.setHeight(y);

  //Create Shares
  cShareSize.setWidth (cSecretSize.width() * cSubPixel.width());
  cShareSize.setHeight(cSecretSize.height()* cSubPixel.height());

  for (unsigned i = 0; i < m_uiUser; i++)
  {
    pcShares[i] = new QImage(cShareSize, QImage::Format_RGB32);
  }

  //for every pixel in the file (x and y are used as index from here on)
  for (x=0; x < static_cast<unsigned>(cSecretSize.width()); x++)
  {
    for (y=0; y < static_cast<unsigned>(cSecretSize.height()); y++)
    {
      //translate pixel
      pcPixel = TranslatePixel(pcSecret->pixel(x, y));

      //write Shares
      for (unsigned i = 0; i < m_uiUser; i++)
      {
        WritePixel(pcShares[i], x, y, pcPixel->GetRow(i, iRet), cSubPixel);
      }
    }
  }

  /******************** write alle shares to harddisk  **********************/
  QString cBase(m_sBaseName.c_str());
  QString cFileName;
  if (cBase=="")
  { //construct Basename, if none has been set
    int iPos;
    QString tempString(m_sSourceFile.c_str());
    iPos = tempString.lastIndexOf("/");
    if (iPos == -1)
      cBase = "/Share";
    else
      cBase = tempString.left(iPos+1)+"Share";
  }

  for (unsigned i = 0; i < m_uiUser; i++)
  {
    cFileName = cBase+QString::number(i+1)+".bmp";
    pcShares[i]->save(cFileName, "BMP", 100);
    (*slRet) << cFileName;
    delete pcShares[i];
  }

  delete[] pcShares;

  return slRet;
}


//This function has to extend the base matrices, so that thay have a
//subpixel expansion of m=x*y afterwards
int VCryptScheme::ModifyBaseMatrices(unsigned x, unsigned y, DMatrix* S0, DMatrix* S1)
{
  int iAddColumns;
  Column* pcZCol;

  //compute number of additional Columns
  iAddColumns = x*y;
  iAddColumns-= S0->GetXDim();

  if (iAddColumns < 0)
    return 0;

  //Create Column that contains only ones
  //(using ones leads to a perfect reconstruction of black pixels)
  pcZCol = new Column(m_uiUser, 1);

  //add Column to base matrices
  while(iAddColumns > 0)
  {
    S0->InsertCol(0, *pcZCol);
    S1->InsertCol(0, *pcZCol);
    iAddColumns--;
  }

  //clean up
  delete pcZCol;

  return 1;
}

//this function has to determine the "optimal" subpixelexpansion that is
//as near at square as possible
void VCryptScheme::FindOptimalExpansion(unsigned m, unsigned &x, unsigned &y,
                                        bool bForceQuadratic)
{
  float fSqrt = sqrt(m);

  if (bForceQuadratic)
    x = y = static_cast<unsigned>(ceil(fSqrt));
  else
  {
    x = static_cast<unsigned>(ceil(fSqrt));
    y = static_cast<unsigned>(floor(fSqrt));

    //if x and y are too small, than increase y by one
    //that suffices for sure, because x=y => x*x >= m
    if ((x*y) < m)
      y++;
  }
}

bool VCryptScheme::WritePixel(QImage* pcImage, unsigned x, unsigned y,
                              Row cPixel, QSize cSubPixelSize)
{
  //Create colors
  QRgb cWhite = qRgb(255, 255, 255);
  QRgb cBlack = qRgb(0, 0, 0);
  //Create startpoint in the share
  unsigned sx = x*cSubPixelSize.width();
  unsigned sy = y*cSubPixelSize.height();

  //Determine
  for (unsigned i = 0; i < cPixel.size(); i++)
  {
    if (cPixel[i]==1)  //Schwarz
      pcImage->setPixel(sx + (i%cSubPixelSize.width()),
                        sy + (i/cSubPixelSize.width()), cBlack);
    else                //Weiss
      pcImage->setPixel(sx + (i%cSubPixelSize.width()),
                        sy + (i/cSubPixelSize.width()), cWhite);
  }

  return true;
}

string VCryptScheme::GetBaseName()
{
  return m_sBaseName;
}

bool VCryptScheme::SetBaseName(string sBaseName)
{
  m_sBaseName = sBaseName;

  return true;
}

string VCryptScheme::GetSecretFileName()
{
  return m_sSourceFile;
}

bool VCryptScheme::SetSecretFileName(string sSourceFile)
{
  m_sSourceFile = sSourceFile;
  return true;
}

//the components of r2 are "ored" on the components of r1
//the resulting Row has the size of r1
Row VCryptScheme::OrRows(const Row& r1, const Row& r2)
{
  Row::iterator i1;
  Row::const_iterator i2;
  Row result=r1;

  i1=result.begin();
  i2=r2.begin();

  while ((i1 != result.end()) && (i2 != r2.end()))
  {
    *i1 = *i1 | *i2;
    ++i1;
    ++i2;
  }

  return result;
}

Row VCryptScheme::OrRows(const set<Row>& sRows)
{
  Row result;
  set<Row>::iterator i = sRows.begin();

  result=*i;
  ++i;
  while(i != sRows.end())
  {
    result = OrRows(*i, result);
    ++i;
  }

  return result;
}

bool VCryptScheme::MatchCols(DMatrix* S0, unsigned uiCol0, DMatrix* S1,
                             unsigned uiCol1, unsigned uiMatchTo)
{
  int iEl1;
  int iEl2;

  //Data valid
  if (S0 == NULL || S1 == NULL)
    return false;

  //Index out of bounds?
  if (S0->GetXDim() <= uiCol0 || S1->GetXDim() <= uiCol1)
    return false;

  //Try to match up to Index uiMatchTo
  for (unsigned i = 0;i < uiMatchTo; i++)
  {
    S0->GetElem(uiCol0, i, iEl1);
    S1->GetElem(uiCol1, i, iEl2);
    if (iEl1 != iEl2)
      return false;  //not matched
  }

  //matched succesfully
  return true;
}

void VCryptScheme::Optimize()
{
  if (m_pcBlackBase == NULL || m_pcWhiteBase == NULL)
    return;

  //find shares that are present in both base matrices
  //and delete them
  //remove columns that are present in S0 and in S1
  unsigned i0 = 0;
  unsigned i1 = 0;
  while(i0 < m_pcWhiteBase->GetXDim())
  {
    if (MatchCols(m_pcWhiteBase, i0, m_pcBlackBase, i1, m_uiUser))
    {
      m_pcWhiteBase->RemoveCol(i0);
      m_pcBlackBase->RemoveCol(i1);
      //test overflow of i1 - overflow of i0 is tested in the outer loop
      if (i1 >= m_pcBlackBase->GetXDim())
      {
        i0++;
        i1 = i0;
      }
    }
    else
    {
      i1++;
      if (i1 >= m_pcBlackBase->GetXDim())
      {
        i0++;
        i1 = i0;
      }
    }
  }
}